from boto3 import client
from builtins import divmod
import logging

logger = logging.getLogger()


class ELBListenerRule:
    def __init__(self, elb_rule_arn: str) -> None:
        self.elb_rule_arn = elb_rule_arn
        self.forward_configs = dict()

    def add_forward_config(self, target_group_arn: str, weight: int) -> None:
        self.forward_configs[target_group_arn] = weight
        return

    def is_restorable(self, source_group_arn: str) -> bool:
        for key in self.forward_configs.keys():
            if key == source_group_arn:
                continue

            if self.forward_configs.get(key) > 0:
                return True

        return False

    def restore(self, source_group_arn: str, weight: int) -> None:
        remaining_weight = weight

        for key in self.forward_configs.keys():
            if key == source_group_arn:
                continue

            weight_to_shed = remaining_weight
            current_weight = self.forward_configs.get(key)

            # restore maximum weight possible from each target.
            if (current_weight < remaining_weight):
                weight_to_shed = current_weight

            self.forward_configs[key] = current_weight - weight_to_shed
            remaining_weight -= weight_to_shed

        self.forward_configs[source_group_arn] = self.forward_configs.get(
            source_group_arn) + weight - remaining_weight

        return

    def shed(self, source_group_arn: str, weight_to_shed: int) -> None:
        source_weight = self.forward_configs.get(source_group_arn)

        if source_weight < weight_to_shed:
            # source weight cannot be < 0
            weight_to_shed = source_weight

        self.forward_configs[source_group_arn] = source_weight - weight_to_shed

        per_target_weight_to_shed = weight_to_shed

        num_forwards = len(self.forward_configs)
        remainder_weight = 0

        if num_forwards > 2:
            # if more than 2 forward configs, then split the weight evenly among the remaining 2 targets
            div_mod = divmod(weight_to_shed, (len(self.forward_configs) - 1))
            per_target_weight_to_shed = div_mod[0]
            remainder_weight = div_mod[1]

        for key in self.forward_configs.keys():
            num_forwards -= 1
            if key == source_group_arn:
                continue

            new_weight = self.forward_configs.get(
                key) + per_target_weight_to_shed

            if (num_forwards == 0):
                new_weight += remainder_weight

            self.forward_configs[key] = new_weight

        return

    def get_target_groups(self) -> list:
        target_groups_list = []

        for key in self.forward_configs.keys():
            target_groups_list.append(
                {'TargetGroupArn': key,
                    'Weight': self.forward_configs.get(key)}
            )

        return target_groups_list

    def save(self, elbv2_client: client):
        elbv2_client.modify_rule(
            RuleArn=self.elb_rule_arn,
            Actions=[
                {
                    'Type': 'forward',
                    'ForwardConfig': {
                        'TargetGroups': self.get_target_groups()
                    }
                }]
        )
